@@ -111,12 +111,10 @@ miner_applications_initable_init (GInitable     *initable,
                                   GCancellable  *cancellable,
                                   GError       **error)
 {
-	TrackerMinerApplications *ma;
 	TrackerMinerFS *fs;
 	GFile *file;
 	GError *inner_error = NULL;
 
-	ma = TRACKER_MINER_APPLICATIONS (initable);
 	fs = TRACKER_MINER_FS (initable);
 
 	/* Chain up parent's initable callback before calling child's one */
@@ -869,26 +867,28 @@ miner_applications_reset (TrackerMiner *miner)
 	g_object_unref (sparql);
 }
 
-TrackerMiner *
-tracker_miner_applications_new (GError **error)
+gboolean
+tracker_miner_applications_detect_locale_changed (TrackerMiner *miner)
 {
-	TrackerMiner *miner;
+	gboolean changed;
 
-	miner = g_initable_new (TRACKER_TYPE_MINER_APPLICATIONS,
-	                        NULL,
-	                        error,
-	                        "name", "Applications",
-	                        "processing-pool-wait-limit", 10,
-	                        "processing-pool-ready-limit", 100,
-	                        NULL);
-	if (miner &&
-	    tracker_miner_applications_locale_changed ()) {
-		/* Before returning the newly created miner, check if we need
-		 * to reset it */
+	changed = tracker_miner_applications_locale_changed ();
+	if (changed) {
 		g_message ("Locale change detected, so resetting miner to "
 		           "remove all previously created items...");
 		miner_applications_reset (miner);
 	}
+	return changed;
+}
 
-	return miner;
+TrackerMiner *
+tracker_miner_applications_new (GError **error)
+{
+	return g_initable_new (TRACKER_TYPE_MINER_APPLICATIONS,
+	                       NULL,
+	                       error,
+	                       "name", "Applications",
+	                       "processing-pool-wait-limit", 10,
+	                       "processing-pool-ready-limit", 100,
+	                       NULL);
 }
