@@ -55,8 +55,8 @@ static TrackerExtractData data[] = {
 };
 
 typedef struct {
-    unsigned int   byteCount;
-    char          *bytes;
+	unsigned int   byteCount;
+	char          *bytes;
 } ExtBlock;
 
 
@@ -67,8 +67,9 @@ ext_block_append(ExtBlock *extBlock,
 		 unsigned char extData[])
 {
 	extBlock->bytes = realloc(extBlock->bytes,extBlock->byteCount+len);
-	if (extBlock->bytes == NULL)
+	if (extBlock->bytes == NULL) {
 		return (GIF_ERROR);
+	}
 
 	memcpy(&(extBlock->bytes[extBlock->byteCount]), &extData[0], len);
 	extBlock->byteCount += len;
@@ -82,25 +83,30 @@ read_metadata (TrackerSparqlBuilder *preupdate,
 	       GifFileType          *gifFile,
                const gchar          *uri)
 {
-	GifRecordType  RecordType;
-	int            frameheight;
-	int            framewidth;
+	GifRecordType RecordType;
+	int frameheight;
+	int framewidth;
 	unsigned char *framedata = NULL;
-	GPtrArray     *keywords;
-	guint          i;
-	int            status;
+	GPtrArray *keywords;
+	guint i;
+	int status;
 	MergeData md = { 0 };
 	GifData   gd = { 0 };
 	TrackerXmpData *xd = NULL;
+	GString *where = NULL;
 
 	do {
+		GifByteType *ExtData;
+		int ExtCode;
+		ExtBlock extBlock;
+
 		if (DGifGetRecordType(gifFile, &RecordType) == GIF_ERROR) {
 			PrintGifError();
 			return;
 		}
 
 		switch (RecordType) {
-		case IMAGE_DESC_RECORD_TYPE:
+			case IMAGE_DESC_RECORD_TYPE:
 			if (DGifGetImageDesc(gifFile) == GIF_ERROR) {
 				PrintGifError();
 				return;
@@ -122,76 +128,70 @@ read_metadata (TrackerSparqlBuilder *preupdate,
 
 			g_free (framedata);
 
-			break;
+		break;
 		case EXTENSION_RECORD_TYPE:
-			{
-				GifByteType   *ExtData;
-				int            ExtCode;
-				ExtBlock       extBlock;
+			extBlock.bytes = NULL;
+			extBlock.byteCount = 0;
 
-				extBlock.bytes = NULL;
-				extBlock.byteCount = 0;
-
-				if ((status = DGifGetExtension (gifFile, &ExtCode, &ExtData)) != GIF_OK) {
-					g_warning ("Problem getting the extension");
-					return;
-				}
+			if ((status = DGifGetExtension (gifFile, &ExtCode, &ExtData)) != GIF_OK) {
+				g_warning ("Problem getting the extension");
+				return;
+			}
 #if defined(HAVE_EXEMPI)
-				if (ExtData && *ExtData &&
-				    strncmp (&ExtData[1],"XMP Data",8) == 0) {
-					while (ExtData != NULL && status == GIF_OK ) {
-						if ((status = DGifGetExtensionNext (gifFile, &ExtData)) == GIF_OK) {
-							if (ExtData != NULL) {
-								if (ext_block_append (&extBlock, ExtData[0]+1, (char *) &(ExtData[0])) != GIF_OK) {
-									g_warning ("Problem with extension data");
-									return;
-								}
+			if (ExtData && *ExtData &&
+			    strncmp (&ExtData[1],"XMP Data",8) == 0) {
+				while (ExtData != NULL && status == GIF_OK ) {
+					if ((status = DGifGetExtensionNext (gifFile, &ExtData)) == GIF_OK) {
+						if (ExtData != NULL) {
+							if (ext_block_append (&extBlock, ExtData[0]+1, (char *) &(ExtData[0])) != GIF_OK) {
+								g_warning ("Problem with extension data");
+								return;
 							}
 						}
 					}
-
-					xd = tracker_xmp_new (extBlock.bytes,
-							      extBlock.byteCount-XMP_MAGIC_TRAILER_LENGTH,
-							      uri);
-
-					g_free (extBlock.bytes);
-				} else
-#endif
-				/* See Section 24. Comment Extension. in the GIF format definition */
-				if (ExtCode == EXTENSION_RECORD_COMMENT_BLOCK_CODE &&
-				    ExtData && *ExtData) {
-					guint block_count = 0;
-
-					/* Merge all blocks */
-					do {
-						block_count++;
-
-						g_debug ("Comment Extension block found (#%u, %u bytes)",
-						         block_count,
-						         ExtData[0]);
-						if (ext_block_append (&extBlock, ExtData[0], (char *) &(ExtData[1])) != GIF_OK) {
-							g_warning ("Problem with Comment extension data");
-							return;
-						}
-					} while (((status = DGifGetExtensionNext(gifFile, &ExtData)) == GIF_OK) &&
-					         ExtData != NULL);
-
-					/* Add last NUL byte */
-					g_debug ("Comment Extension blocks found (%u) with %u bytes",
-					         block_count,
-					         extBlock.byteCount);
-					extBlock.bytes = g_realloc (extBlock.bytes, extBlock.byteCount + 1);
-					extBlock.bytes[extBlock.byteCount] = '\0';
-
-					/* Set commentt */
-					gd.comment = extBlock.bytes;
-				} else {
-					do {
-						status = DGifGetExtensionNext(gifFile, &ExtData);
-					} while ( status == GIF_OK && ExtData != NULL);
 				}
+
+				xd = tracker_xmp_new (extBlock.bytes,
+				                      extBlock.byteCount-XMP_MAGIC_TRAILER_LENGTH,
+				                      uri);
+
+				g_free (extBlock.bytes);
+			} else
+#endif
+			/* See Section 24. Comment Extension. in the GIF format definition */
+			if (ExtCode == EXTENSION_RECORD_COMMENT_BLOCK_CODE &&
+			    ExtData && *ExtData) {
+				guint block_count = 0;
+
+				/* Merge all blocks */
+				do {
+					block_count++;
+
+					g_debug ("Comment Extension block found (#%u, %u bytes)",
+					         block_count,
+					         ExtData[0]);
+					if (ext_block_append (&extBlock, ExtData[0], (char *) &(ExtData[1])) != GIF_OK) {
+						g_warning ("Problem with Comment extension data");
+						return;
+					}
+				} while (((status = DGifGetExtensionNext(gifFile, &ExtData)) == GIF_OK) &&
+				         ExtData != NULL);
+
+				/* Add last NUL byte */
+				g_debug ("Comment Extension blocks found (%u) with %u bytes",
+				         block_count,
+				         extBlock.byteCount);
+				extBlock.bytes = g_realloc (extBlock.bytes, extBlock.byteCount + 1);
+				extBlock.bytes[extBlock.byteCount] = '\0';
+
+				/* Set commentt */
+				gd.comment = extBlock.bytes;
+			} else {
+				do {
+					status = DGifGetExtensionNext(gifFile, &ExtData);
+				} while ( status == GIF_OK && ExtData != NULL);
 			}
-			break;
+		break;
 		case TERMINATE_RECORD_TYPE:
 			break;
 		default:
@@ -351,18 +351,35 @@ read_metadata (TrackerSparqlBuilder *preupdate,
 	}
 
 	for (i = 0; i < keywords->len; i++) {
-		gchar *p;
+		gchar *p, *escaped, *var;
 
 		p = g_ptr_array_index (keywords, i);
+		escaped = tracker_sparql_escape_string (p);
+		var = g_strdup_printf ("tag%d", i + 1);
 
+		/* ensure tag with specified label exists */
+		tracker_sparql_builder_append (preupdate,
+		                               "INSERT { _:tag a nao:Tag ; nao:prefLabel \"");
+		tracker_sparql_builder_append (preupdate, escaped);
+		tracker_sparql_builder_append (preupdate,
+		                               "\" }\nWHERE { FILTER (NOT EXISTS { "
+		                               "?tag a nao:Tag ; nao:prefLabel \"");
+		tracker_sparql_builder_append (preupdate, escaped);
+		tracker_sparql_builder_append (preupdate,
+		                               "\" }) }\n");
+
+		/* associate file with tag */
 		tracker_sparql_builder_predicate (metadata, "nao:hasTag");
+		tracker_sparql_builder_object_variable (metadata, var);
 
-		tracker_sparql_builder_object_blank_open (metadata);
-		tracker_sparql_builder_predicate (metadata, "a");
-		tracker_sparql_builder_object (metadata, "nao:Tag");
-		tracker_sparql_builder_predicate (metadata, "nao:prefLabel");
-		tracker_sparql_builder_object_unvalidated (metadata, p);
-		tracker_sparql_builder_object_blank_close (metadata);
+		if (where == NULL) {
+			where = g_string_new ("} } WHERE { {\n");
+		}
+
+		g_string_append_printf (where, "?%s a nao:Tag ; nao:prefLabel \"%s\" .\n", var, escaped);
+
+		g_free (var);
+		g_free (escaped);
 		g_free (p);
 	}
 	g_ptr_array_free (keywords, TRUE);
@@ -418,52 +435,74 @@ read_metadata (TrackerSparqlBuilder *preupdate,
 		tracker_sparql_builder_object_unvalidated (metadata, xd->coverage);
 	}
 
-	if (xd->address || xd->country || xd->city) {
-		gchar *addruri;
+	if (xd->address || xd->state || xd->country || xd->city ||
+	    xd->gps_altitude || xd->gps_latitude || xd-> gps_longitude) {
 
 		tracker_sparql_builder_predicate (metadata, "slo:location");
 
-		tracker_sparql_builder_object_blank_open (metadata); /* GeoPoint */
+		tracker_sparql_builder_object_blank_open (metadata); /* GeoLocation */
 		tracker_sparql_builder_predicate (metadata, "a");
-		tracker_sparql_builder_object (metadata, "slo:GeoPoint");
+		tracker_sparql_builder_object (metadata, "slo:GeoLocation");
 
-		addruri = tracker_sparql_get_uuid_urn ();
+		if (xd->address || xd->state || xd->country || xd->city)  {
+			gchar *addruri;
+			addruri = tracker_sparql_get_uuid_urn ();
 
-		tracker_sparql_builder_predicate (metadata, "slo:postalAddress");
-		tracker_sparql_builder_object_iri (metadata, addruri);
+			tracker_sparql_builder_predicate (metadata, "slo:postalAddress");
+			tracker_sparql_builder_object_iri (metadata, addruri);			
 
-		tracker_sparql_builder_object_blank_close (metadata); /* GeoPoint */
+			tracker_sparql_builder_insert_open (preupdate, NULL);
+			tracker_sparql_builder_subject_iri (preupdate, addruri);
 
-		tracker_sparql_builder_insert_open (preupdate, NULL);
-		tracker_sparql_builder_subject_iri (preupdate, addruri);
+			g_free (addruri);
 
-		g_free (addruri);
+			tracker_sparql_builder_predicate (preupdate, "a");
+			tracker_sparql_builder_object (preupdate, "nco:PostalAddress");
 
-		tracker_sparql_builder_predicate (preupdate, "a");
-		tracker_sparql_builder_object (preupdate, "nco:PostalAddress");
+			if (xd->address) {
+				tracker_sparql_builder_predicate (preupdate, "nco:streetAddress");
+				tracker_sparql_builder_object_unvalidated (preupdate, xd->address);
+			}
 
-		if (xd->address) {
-			tracker_sparql_builder_predicate (preupdate, "nco:streetAddress");
-			tracker_sparql_builder_object_unvalidated (preupdate, xd->address);
+			if (xd->state) {
+				tracker_sparql_builder_predicate (preupdate, "nco:region");
+				tracker_sparql_builder_object_unvalidated (preupdate, xd->state);
+			}
+
+			if (xd->city) {
+				tracker_sparql_builder_predicate (preupdate, "nco:locality");
+				tracker_sparql_builder_object_unvalidated (preupdate, xd->city);
+			}
+
+			if (xd->country) {
+				tracker_sparql_builder_predicate (preupdate, "nco:country");
+				tracker_sparql_builder_object_unvalidated (preupdate, xd->country);
+			}
+
+			tracker_sparql_builder_insert_close (preupdate);
 		}
 
-		if (xd->state) {
-			tracker_sparql_builder_predicate (preupdate, "nco:region");
-			tracker_sparql_builder_object_unvalidated (preupdate, xd->state);
+		if (xd->gps_altitude) {
+			tracker_sparql_builder_predicate (metadata, "slo:altitude");
+			tracker_sparql_builder_object_unvalidated (metadata, xd->gps_altitude);
 		}
 
-		if (xd->city) {
-			tracker_sparql_builder_predicate (preupdate, "nco:locality");
-			tracker_sparql_builder_object_unvalidated (preupdate, xd->city);
+		if (xd->gps_latitude) {
+			tracker_sparql_builder_predicate (metadata, "slo:latitude");
+			tracker_sparql_builder_object_unvalidated (metadata, xd->gps_latitude);
 		}
 
-		if (xd->country) {
-			tracker_sparql_builder_predicate (preupdate, "nco:country");
-			tracker_sparql_builder_object_unvalidated (preupdate, xd->country);
+		if (xd->gps_longitude) {
+			tracker_sparql_builder_predicate (metadata, "slo:longitude");
+			tracker_sparql_builder_object_unvalidated (metadata, xd->gps_longitude);
 		}
 
-		tracker_sparql_builder_insert_close (preupdate);
+		tracker_sparql_builder_object_blank_close (metadata); /* GeoLocation */
+	}
 
+	if (xd->gps_direction) {
+		tracker_sparql_builder_predicate (metadata, "nfo:heading");
+		tracker_sparql_builder_object_unvalidated (metadata, xd->gps_direction);
 	}
 
 	if (gd.width) {
@@ -484,6 +523,11 @@ read_metadata (TrackerSparqlBuilder *preupdate,
 		g_free (gd.comment);
 	}
 
+	if (where != NULL) {
+		tracker_sparql_builder_append (metadata, where->str);
+		g_string_free (where, TRUE);
+	}
+
 	tracker_xmp_free (xd);
 }
 
@@ -493,9 +537,9 @@ extract_gif (const gchar          *uri,
              TrackerSparqlBuilder *preupdate,
              TrackerSparqlBuilder *metadata)
 {
-	goffset      size;
+	goffset size;
 	GifFileType *gifFile = NULL;
-	gchar       *filename;
+	gchar *filename;
 
 	filename = g_filename_from_uri (uri, NULL, NULL);
 	size = tracker_file_get_size (filename);
